@@ -1,13 +1,12 @@
-# $Id: REST.pm 38 2010-03-10 17:30:50Z pplusdomain $
 # RT::Client::REST
 #
 # Dmitri Tikhonov <dtikhonov@yahoo.com>
-# April 18, 2006
 #
 # Part of the source is Copyright (c) 2007-2008 Damien Krotkine <dams@cpan.org>
 #
-# This code is adapted (stolen) from /usr/bin/rt that came with RT.  I just
-# wanted to make an actual module out of it.  Therefore, this code is GPLed.
+# This code is adapted from /usr/bin/rt that came with RT.  As of version 0.49,
+# this module is licensed using Perl Artistic License, with permission from the
+# original author of rt utility, Abhijit Menon-Sen.
 #
 # Original notice:
 #------------------------
@@ -25,7 +24,8 @@ use strict;
 use warnings;
 
 use vars qw/$VERSION/;
-$VERSION = 0.40;
+$VERSION = '0.50';
+$VERSION = eval $VERSION;
 
 use Error qw(:try);
 use HTTP::Cookies;
@@ -39,7 +39,14 @@ for my $method (qw(server _cookie timeout)) {
     no strict 'refs';
     *{__PACKAGE__ . '::' . $method} = sub {
         my $self = shift;
-        $self->{'_' . $method} = shift if @_;
+        if (@_) {
+            my $val = shift;
+            {
+                no warnings 'uninitialized';
+                $self->logger->debug("set `$method' to $val");
+            }
+            $self->{'_' . $method} = $val;
+        }
         return $self->{'_' . $method};
     };
 }
@@ -49,10 +56,16 @@ sub new {
 
     $class->_assert_even(@_);
 
-    my $self = bless {}, ref($class) || $class;
+    my $self = bless {
+        _logger => RT::Client::REST::NoopLogger->new,
+    }, ref($class) || $class;
     my %opts = @_;
 
     while (my ($k, $v) = each(%opts)) {
+        # in _rest we concatenate server with '/REST/1.0';
+        if ($k eq 'server') {
+            $v =~ s!/$!!;
+        }
         $self->$k($v);
     }
 
@@ -65,21 +78,21 @@ sub login {
     $self->_assert_even(@_);
 
     my %opts = @_;
-    unless (defined($opts{username}) and defined($opts{password})) {
+    unless (scalar(keys %opts) > 0) {
         RT::Client::REST::InvalidParameterValueException->throw(
-            "You must provide username and password to log in",
+            "You must provide credentials (user and pass) to log in",
         );
     }
+    # back-compat hack
+    if (defined $opts{username}){ $opts{user} = $opts{username}; delete $opts{username} }
+    if (defined $opts{password}){ $opts{pass} = $opts{password}; delete $opts{password} }
 
     # OK, here's how login works.  We request to see ticket 1.  We don't
     # even care if it exists.  We watch exceptions: auth. failures and
     # server-side errors we bubble up and ignore all others.
     try {
         $self->_cookie(undef);  # Start a new session.
-        $self->_submit("ticket/1", undef, {
-            user => $opts{username},
-            pass => $opts{password},
-        });
+        $self->_submit("ticket/1", undef, \%opts);
     } catch RT::Client::REST::AuthenticationFailureException with {
         shift->rethrow;
     } catch RT::Client::REST::MalformedRTResponseException with {
@@ -93,7 +106,6 @@ sub login {
     };
 }
 
-
 sub show {
     my $self = shift;
 
@@ -104,7 +116,7 @@ sub show {
     my $type = $self->_valid_type(delete($opts{type}));
     my $id;
 
-    if (grep { $type eq $_ } (qw(user queue))) {
+    if (grep { $type eq $_ } (qw(user queue group))) {
         # User or queue ID does not have to be numeric
         $id = delete($opts{id});
     } else {
@@ -154,8 +166,37 @@ sub get_attachment {
     my $parent_id = $self->_valid_numeric_object_id(delete($opts{parent_id}));
     my $id = $self->_valid_numeric_object_id(delete($opts{id}));
 
+    my $res = $self->_submit("$type/$parent_id/attachments/$id");
+    my $content;
+    if ($opts{undecoded}) {
+        $content = $res->content;
+    }
+    else {
+        $content = $res->decoded_content;
+    }
+    my $form = form_parse($content);
+
+    my ($c, $o, $k, $e) = @{$$form[0]};
+
+    if (!@$o && $c) {
+        RT::Client::REST::Exception->_rt_content_to_exception($c)->throw;
+    }
+
+    return $k;
+}
+
+sub get_links {
+    my $self = shift;
+
+    $self->_assert_even(@_);
+
+    my %opts = @_;
+
+    my $type = $self->_valid_type(delete($opts{type}) || 'ticket');
+    my $id = $self->_valid_numeric_object_id(delete($opts{id}));
+
     my $form = form_parse(
-        $self->_submit("$type/$parent_id/attachments/$id")->decoded_content
+        $self->_submit("$type/$id/links/$id")->decoded_content
     );
     my ($c, $o, $k, $e) = @{$$form[0]};
 
@@ -163,6 +204,31 @@ sub get_attachment {
         RT::Client::REST::Exception->_rt_content_to_exception($c)->throw;
     }
 
+    # Turn the links into id lists
+    foreach my $key (keys(%$k)) {
+        try {
+            $self->_valid_link_type($key);
+            my @list = split(/\s*,\s*/,$k->{$key});
+            #use Data::Dumper;
+            #print STDERR Dumper(\@list);
+            my @newlist = ();
+            foreach my $val (@list) {
+               if ($val =~ /^fsck\.com-\w+\:\/\/(.*?)\/(.*?)\/(\d+)$/) {
+                   # We just want the ids, not the URI
+                   push(@newlist, {'type' => $2, 'instance' => $1, 'id' => $3 });
+               } else {
+                   # Something we don't recognise
+                   push(@newlist, { 'url' => $val });
+               }
+            }
+            # Copy the newly created list
+            $k->{$key} = ();
+            $k->{$key} = \@newlist;
+        } catch RT::Client::REST::InvalidParameterValueException with {
+            # Skip it because the keys are not always valid e.g., 'id'
+        }
+    }
+
     return $k;
 }
 
@@ -370,11 +436,11 @@ sub merge_tickets {
     my %opts = @_;
     my ($src, $dst) = map { $self->_valid_numeric_object_id($_) }
         @opts{qw(src dst)};
-    $self->_submit("ticket/merge/$src", { into => $dst});
+    $self->_submit("ticket/$src/merge/$dst");
     return;
 }
 
-sub link_tickets {
+sub link {
     my $self = shift;
     $self->_assert_even(@_);
     my %opts = @_;
@@ -382,8 +448,13 @@ sub link_tickets {
         @opts{qw(src dst)};
     my $ltype = $self->_valid_link_type(delete($opts{link_type}));
     my $del = (exists($opts{'unlink'}) ? 1 : '');
+    my $type = $self->_valid_type(delete($opts{type}) || 'ticket');
 
-    $self->_submit("ticket/link", {
+    #$self->_submit("$type/$src/link", {
+    #id => $from, rel => $rel, to => $to, del => $del
+    #}
+
+    $self->_submit("$type/link", {
         id  => $src,
         rel => $ltype,
         to  => $dst,
@@ -393,7 +464,10 @@ sub link_tickets {
     return;
 }
 
-sub unlink_tickets { shift->link_tickets(@_, unlink => 1) }
+sub link_tickets { shift->link(@_, type => 'ticket') }
+
+sub unlink { shift->link(@_, unlink => 1) }
+sub unlink_tickets { shift->link(@_, type => 'ticket', unlink => 1) }
 
 sub _ticket_action {
     my $self = shift;
@@ -460,6 +534,9 @@ sub _submit {
 
     # Now, we construct the request.
     if (@$data) {
+        # The request object expects "bytes", not strings
+        map { utf8::encode($_) unless ref($_)} @$data;
+
         $req = POST($self->_uri($uri), $data, Content_Type => 'form-data');
     }
     else {
@@ -471,9 +548,9 @@ sub _submit {
     }
 
     # Then we send the request and parse the response.
-    #DEBUG(3, $req->as_string);
+    $self->logger->debug("request: ", $req->as_string);
     my $res = $self->_ua->request($req);
-    #DEBUG(3, $res->as_string);
+    $self->logger->debug("response: ", $res->as_string);
 
     if ($res->is_success) {
         # The content of the response we get from the RT server consists
@@ -540,6 +617,17 @@ sub _submit {
         RT::Client::REST::RequestTimedOutException->throw(
             "Your request to " . $self->server . " timed out",
         );
+    } elsif (302 == $res->code && !$self->{'_redirected'}) {
+        $self->{'_redirected'} = 1;     # We only allow one redirection
+        # Figure out the new value of 'server'.  We assume that the /REST/..
+        # part of the URI stays the same.
+        my $new_location = $res->header('Location');
+        $self->logger->info("We're being redirected to $new_location");
+        my $orig_server = $self->server;
+        (my $suffix = $self->_uri($uri)) =~ s/^\Q$orig_server//;
+        (my $new_server = $new_location) =~ s/\Q$suffix\E$//;
+        $self->server($new_server);
+        return $self->_submit($uri, $content, $auth);
     } else {
         RT::Client::REST::HTTPException->throw(
             code    => $res->code,
@@ -585,6 +673,25 @@ sub basic_auth_cb {
     return $self->{_basic_auth_cb};
 }
 
+use constant LOGGER_METHODS => (qw(debug warn info error));
+
+sub logger {
+    my $self = shift;
+    if (@_) {
+        my $new_logger = shift;
+        for my $method (LOGGER_METHODS) {
+            unless ($new_logger->can($method)) {
+                RT::Client::REST::InvalidParameterValueException->throw(
+                    "logger does not know how to `$method'",
+                );
+            }
+        }
+        $self->{'_logger'} = $new_logger;
+    }
+    return $self->{'_logger'};
+}
+
+
 # Not a constant so that it can be overridden.
 sub _list_of_valid_transaction_types {
     sort +(qw(
@@ -655,8 +762,8 @@ sub _valid_comment_message {
 
 sub _valid_link_type {
     my ($self, $type) = @_;
-    my @types = qw(DependsOn DependedOnBy RefersTo ReferredToBy HasMember
-                   MemberOf);
+    my @types = qw(DependsOn DependedOnBy RefersTo ReferredToBy HasMember Members
+                   MemberOf RunsOn IsRunning ComponentOf HasComponent);
 
     unless (grep { lc($type) eq lc($_) } @types) {
         RT::Client::REST::InvalidParameterValueException->throw(
@@ -708,6 +815,39 @@ sub _ua_string {
 
 sub _version { $VERSION }
 
+{
+    # This is a noop logger: it discards all log messages.  It is the default
+    # logger.  I think this approach is better than doing either checks all
+    # over the place like this:
+    #
+    #   if ($self->logger) {
+    #       $self->logger->warn("message");
+    #   }
+    #
+    # or creating our own logging methods which will hide the checks:
+    #
+    #   sub warn {
+    #       my $self = shift;
+    #       if ($self->logger) {
+    #           $self->logger->warn(@_);
+    #       }
+    #   }
+    #   # and later:
+    #   sub xyz {
+    #       ...
+    #       $self->warn("message");
+    #   }
+    #
+    # The problem with the second approach is that it creates unrelated
+    # methods in RT::Client::REST namespace.
+    package RT::Client::REST::NoopLogger;
+    sub new { bless \(my $logger) }
+    for my $method (RT::Client::REST::LOGGER_METHODS) {
+        no strict 'refs';
+        *{$method} = sub {};
+    }
+}
+
 1;
 
 __END__
@@ -780,7 +920,8 @@ of type B<HTTP::Cookies> to use for credentials information.
 
 B<timeout> is the number of seconds HTTP client will wait for the
 server to respond.  Defaults to LWP::UserAgent's default timeout, which
-is 300 seconds.
+is 180 seconds (please check LWP::UserAgent's documentation for accurate
+timeout information).
 
 =item B<basic_auth_cb>
 
@@ -795,9 +936,16 @@ returns username and password:
     return ($username, $password);
   }
 
+=item B<logger>
+
+A logger object.  It should be able to debug(), info(), warn() and
+error().  It is not widely used in the code (yet), and so it is
+mostly useful for development.
+
 =back
 
 =item login (username => 'root', password => 'password')
+=item login (my_userfield => 'root', my_passfield => 'password')
 
 Log in to RT.  Throws an exception on error.
 
@@ -808,7 +956,8 @@ See B<basic_auth_cb> above.
 =item show (type => $type, id => $id)
 
 Return a reference to a hash with key-value pair specifying object C<$id>
-of type C<$type>.
+of type C<$type>. The keys are the names of RT's fields. Keys for custom
+fields are in the form of "CF.{CUST_FIELD_NAME}".
 
 =item edit (type => $type, id => $id, set => { status => 1 })
 
@@ -862,8 +1011,8 @@ using C<show()> method:
     query => "Status = 'stalled'",
   );
   for my $id (@ids) {
-    my ($ticket) = $rt->show(type => 'ticket', ids => [$id]);
-    print "Subject: ", $t->{Subject}, "\n";
+    my ($ticket) = $rt->show(type => 'ticket', id => $id);
+    print "Subject: ", $ticket->{Subject}, "\n";
   }
 
 =item comment (ticket_id => $id, message => $message, %opts)
@@ -888,12 +1037,21 @@ B<bcc>, and B<attachments> parameters (see C<comment> above).
 
 Get a list of numeric attachment IDs associated with ticket C<$id>.
 
-=item get_attachment (parent_id => $parent_id, id => $id)
+=item get_attachment (parent_id => $parent_id, id => $id, undecoded => $bool)
 
 Returns reference to a hash with key-value pair describing attachment
 C<$id> of ticket C<$parent_id>.  (parent_id because -- who knows? --
 maybe attachments won't be just for tickets anymore in the future).
 
+If the option undecoded is set to a true value, the attachment will be
+returned verbatim and undecoded (this is probably what you want with
+images and binary data).
+
+=item get_links (type =E<gt> $type, id =E<gt> $id)
+
+Get link information for object of type $type whose id is $id.
+If type is not specified, 'ticket' is used.
+
 =item get_transaction_ids (parent_id => $id, %opts)
 
 Get a list of numeric IDs associated with parent ID C<$id>.  C<%opts>
@@ -1080,7 +1238,7 @@ RT server, which is either good or bad, depending how you look at it.
 
 =head1 VERSION
 
-This is version 0.40 of B<RT::Client::REST>.
+This is version 0.50 of B<RT::Client::REST>.
 
 =head1 AUTHORS
 
@@ -1088,9 +1246,10 @@ Original /usr/bin/rt was written by Abhijit Menon-Sen <ams@wiw.org>.  rt
 was later converted to this module by Dmitri Tikhonov <dtikhonov@yahoo.com>.
 In January of 2008, Damien "dams" Krotkine <dams@cpan.org> joined as the
 project's co-maintainer. JLMARTIN has become co-maintainer as of March 2010.
+SRVSH became a co-maintainer in November 2015.
 
 =head1 LICENSE
 
-Since original rt is licensed under GPL, so is this module.
+Perl license.
 
 =cut
